// antmh.cc -- Metsahovi serial antenna interface, for FS9.
// Copyright (C) 1995 Ari Mujunen. (Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log: antmh.cc,v $
// Revision 1.1  1997/03/13  12:30:15  prog
// Initial revision
//


#include "registration.h"
RCSID(antmh_cc, "$Id: antmh.cc,v 1.1 1997/03/13 12:30:15 prog Exp prog $");

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "cSerialLine.hh"

#include "antmh.hh"


/* Adjustable behavior. */
#define MAX_COMMS_RETRIES 3

/* Module static/instance variables. */
tAntennaStatus vAntennaStatusReturnedByLatestReceiveFromVAX = cStStandby;
double vTrackRa = -1.0;  /* not tracking */
double vTrackDecl = 999.0;
double vRaOff = 0.0;
double vDeclOff = 0.0;
double vAzOff = 0.0;
double vElOff = 0.0;
class cSerialLine vLine;

/* Internal routines. */

static inline double
fToDeg(double d) {
  return (d/M_PI*180.0);
}

  /* Converts right ascension ra (in radians) into a string of form:  hhmmss.s
  ra must be in the range [0,2pi) and str has to have room for 9 characters.*/
static void
pRAToStr(
  double ra,
  char *str
) {
  double hh;
  double mm;
  double ss;

  ra = fToDeg(ra);

  ra = ra/360.0*24.0;
  hh = floor(ra);
  ra = 60.0*(ra-hh);
  mm = floor(ra);
  ss = 60.0*(ra-mm);

  /* If 'ss' will round, we need to carry over into 'mm' and 'hh'. */
  if (ss >= 59.95) {
    ss = 0.0; mm = mm + 1.0;
    if (mm > 59) {
      mm = 0.0; hh = hh + 1.0;
      if (hh > 23.0) {
        hh = 0.0;
      }
    }
  }

  sprintf(str, "%02.0f%02.0f%04.1f", hh, mm, ss);
}  /* pRAToStr */

  /* Converts declination decl (in radians) into a string of form:  Sddmmss.s
  decl must be in the range [-pi/2,pi/2] 
  and 'str' has to have room for 10 characters. */
static void
pDeclToStr(
  double decl,
  char * str
) {
  double dd;
  double mm;
  double ss;
  char signum = '+';

  if (decl < 0) {
    decl = -decl;
    signum = '-';
  }

  decl = fToDeg(decl);

  dd = floor(decl); decl = 60.0*(decl-dd);
  mm = floor(decl);
  ss = 60.0*(decl-mm);

  /* If 'ss' will round, we need to carry over into 'mm' and 'hh'. */
  if (ss >= 59.95) {
    ss = 0.0; mm = mm + 1.0;
    if (mm > 59.0) {
      mm = 0.0; dd = dd + 1.0;
    }
  }

  sprintf(str, "%c%02.0f%02.0f%04.1f", signum, dd, mm, ss);
}  /* pDeclToStr */

/* Sends a command to VAX VLBI_TRACKER. */
static void
pSendToVax(
  double ra,
  double decl,
  double az_off,
  double el_off,
  char vFlagCharacter
) {

/*
  Sends source position parameters and a control flag to Vax. Also flushes
  the input buffer, since new response is expected promptly.

  Pars: ra            - right ascension in radians [0,2pi) for current epoch
        decl          - declination in degrees [-pi/2,pi/2]
        az_off,el_off - azimuth and elevation offsets in radians
                        corresponding to degree range (-10,10)
        vFlagCharacter- flag for controlling the antenna steering program
                        (one of cCo* characters).
*/

  char buf[8 +9   +5    +5    +1   +1];
        /* ra+decl+azoff+eloff+flag+\0 */

  pRAToStr(ra, &(buf[0]));
  pDeclToStr(decl, &(buf[8]));
  sprintf(&(buf[17]), "%+05.0f%+05.0f%c",
	  fToDeg(az_off)*1000.0, fToDeg(el_off)*1000.0,
    vFlagCharacter);

  /* Flush incoming characters, send the buffer. */
  vLine.flushPendingInput();
  vLine.write(buf, sizeof(buf)-1);
}  /* pSendToVax */


tAntennaStatus
fReceiveFromVax(char *str) {

/*
  Receives time + status from Vax. Time is stored as a string in str in
  format (Vax): YYYYMMDDHHMMSS.SSF with a constant length of 18.
  str must have space for at least 18 characters, or, it must be NULL.
  In this case, only the status code is returned.

  If successful, returns a status integer code (cSt*), otherwise -1.
*/

  char *buf;
  char tempBuf[18];

  if (str) {
    buf = str;
  } else {
    buf = tempBuf;
  }

  if ((vLine.read(buf, 18)) < 18) {
     return (cStNoResponse);
  }
  if ((int)(buf[17]) > cStLastDefinedCode) {
    /* Responses fromVAX don't obey the protocol. */
    return (cStNoResponse);
  }

  /* If communication protocol is in sync, remember the latest status. */
  vAntennaStatusReturnedByLatestReceiveFromVAX = (enum tAntennaStatus)(buf[17]);

  /* Turn time string into a null-terminated string by overwriting */
  /* status byte with '\0'. */
  buf[17] = 0;

  return (vAntennaStatusReturnedByLatestReceiveFromVAX);
}  /* fReceiveFromVax */


/* Antenna routines. */

  /* Returns antenna status from the latest antenna operation. */
  /* Useful in continuously updating displays etc. */
tAntennaStatus
fPeekAntennaStatus(void) {
  return (vAntennaStatusReturnedByLatestReceiveFromVAX);
}  /* fPeekAntennaStatus */

  /* Gets a "fresh" antenna status from antenna control computer. */
tAntennaStatus
fQueryAntennaStatus(char *stStr) {
  tAntennaStatus status;

  if (vTrackRa > -0.5) {  /* -1.0 signifies not tracking */
    pSendToVax(vTrackRa+vRaOff, vTrackDecl+vDeclOff,
	       vAzOff, vElOff, cCoCheckPointing);
  } else {
    pSendToVax(0.0, 0.0, 0.0, 0.0, cCoSendTime);
  }
  status = fReceiveFromVax(stStr);
  if (status > cStNoResponse) {
    vAntennaStatusReturnedByLatestReceiveFromVAX = status;
  }  /* if responded */
  return (status);
}  /* fQueryAntennaStatus */


char *
fDecodeAntennaStatus(
  tAntennaStatus status,
  int maxLength
  /*int recommendedColor*/
) {
/* Was "variable" in gcc:
  static const int numCodes = cStLastDefinedCode - cStNoResponse + 2;
 */
#define numCodes (cStLastDefinedCode - cStNoResponse + 2)
#define idx(x) ((x)-(cStNoResponse))
#if 0
  static const int cColors[numCodes] = {
    red, magenta, black, brown, blue, red, red, red
  };
#endif
  static const char cLong[numCodes][12] = {
    "NO RESPONSE", "TIME ONLY", "STANDBY", "SLEWING",
    "TRACKING", "LIMITS", "DEV ERROR", "UNKNOWN"
  };
  static const char cMedium[numCodes][6] = {
    "NORSP", "TIME", "STDBY", "SLEW",
    "TRACK", "LIMIT", "DVERR", "UNKNW"
  };
  static const char cShort[numCodes][3] = {
    "NR", "TI", "SB", "SL",
    "TR", "LI", "DV", "UN"
  };

  static char stStr[20];

  if ((status < cStNoResponse) || (status > cStLastDefinedCode)) {
    status = (enum tAntennaStatus)(cStLastDefinedCode+1);
  }

  if (maxLength < 4) 
    strcpy(stStr, cShort[idx(status)]);
  else if (maxLength < 9)
    strcpy(stStr, cMedium[idx(status)]);
  else
    strcpy(stStr, cLong[idx(status)]);
  if (maxLength > 19) 
    maxLength = 19;
  stStr[maxLength] = '\0';

  /*recommendedColor = cColors[idx(status)];*/
 return (stStr);
#undef idx
}  /* fDecodeAntennaStatus */

tAntennaStatus
fGetTimeFromAntennaComputer(
  char *timeString
) {
  tAntennaStatus status;

  pSendToVax(0.0, 0.0, 0.0, 0.0, cCoSendTime);
  status = fReceiveFromVax(timeString);
  if (status > cStNoResponse) {
    vAntennaStatusReturnedByLatestReceiveFromVAX = status;
  }  /* if responded */
  return (status);
}  /* fGetTimeFromAntennaComputer */

/* Start/stop tracking a ra/decl source. */
tAntennaStatus
fTrackASource(
  double ra,
  double decl
) {
  tAntennaStatus status;

  pSendToVax(ra+vRaOff, decl+vDeclOff, vAzOff, vElOff, cCoTrack);
  status = fReceiveFromVax(NULL);
  if (status > cStNoResponse) {
    vTrackRa = ra;
    vTrackDecl = decl;
    vAntennaStatusReturnedByLatestReceiveFromVAX = status;
  }
  return (status);
}  /* fTrackASource */

/* Goto a given az-el coordinate. */
tAntennaStatus
fGotoAzEl(
  double az,
  double el
) {
  tAntennaStatus status;

  pSendToVax(az /*in place of ra (0..360 deg/0..24h) */,
                // VAX transforms from FS 0..360 to its own -180..180.
	     el /*in place of dec (-90..90 deg) */,
	     vAzOff, vElOff,  // VAX actually ignores these.
	                      // (They are processed in pointing model
                              // and az-el bypasses that.)
	     cCoGotoAzEl);
  status = fReceiveFromVax(NULL);
  if (status > cStNoResponse) {
    // As if it were stopped:
    vTrackRa = -1.0;
    vTrackDecl = 999.0;
    vAntennaStatusReturnedByLatestReceiveFromVAX = status;
  }
  return (status);
}  /* fGotoAzEl */

tAntennaStatus
fStopTracking(void) {
  tAntennaStatus status;

  pSendToVax(0.0, 0.0, 0.0, 0.0, cCoStopTracking);
  status = fReceiveFromVax(NULL);
  if (status > cStNoResponse) {
    vTrackRa = -1.0;
    vTrackDecl = 999.0;
    vAntennaStatusReturnedByLatestReceiveFromVAX = status;
  }  /* if responded */
  return (status);
}  /* fStopTracking */


/* FS9 interface routines. */
void
fsAntOpen( char *path) {
  vLine.open(path);
  vLine.serialSettings(9600, 8, 1, eParityNone, 0, /*maxTimeout=>*/ 5);
}  /* fsAntOpen */

int  /* 0==successful */
fsStartTracking(char *lsorna,
		double ra50, double dec50, double ep1950,
		double radat, double decdat, double epoch
) {
  int i;

  /* Retry up to MAX_COMMS_RETRIES times. */
  for (i=0; i<MAX_COMMS_RETRIES; i++) {
    tAntennaStatus res;

    res = fTrackASource(radat, decdat);
      /* xxx: really assume that FS has converted these to today's epoch */

    if (res != cStNoResponse) {
      return 0;
    }
  }  /* for */
  return 1;  /* 'source=' failed */
}  /* fsStartTracking */

int  /* 0==successful */
fsGotoAzEl(
  double az,  // in FS azimuth (0..360 deg) but in radians
  double el
) {
  int i;

  /* Retry up to MAX_COMMS_RETRIES times. */
  for (i=0; i<MAX_COMMS_RETRIES; i++) {
    tAntennaStatus res;

    res = fGotoAzEl(az, el);  // in FS az
    if (res != cStNoResponse) {
      return 0;
    }
  }  /* for */
  return 1;  /* 'source=' failed */
}  /* fsGotoAzEl */

void
fsGotoIdle(void) {
  fStopTracking();
}  /* fsGotoIdle */

void
fsUpdateOffsets(double raOff, double decOff,
		double azOff, double elOff, 
		double xOff, double yOff
) {
  /* Cannot yet handle X/Y offsets. */
  assert( (xOff == 0.0) && (yOff == 0.0) );

  vRaOff = raOff;
  vDeclOff = decOff;
  vAzOff = azOff;
  vElOff = elOff;
  fsStartTracking(NULL, 0.0, 0.0, 0.0,
		  vTrackRa, vTrackDecl, 0.0);
  /* xxx: many parameters missing, since we know they won't be used */
  /* xxx: and if 'idle', vTrackRa = -1.0 rad will get to antenna... :-( */
}  /* fsUpdateOffsets */

/* Returns 1 only iff on source, status strings in '*buf'. */
int
fsOnsource(char *buf, char *verboseBuf) {
  tAntennaStatus st;
  int i;

  /* Retry up to MAX_COMMS_RETRIES times. */
  for (i=0; i<MAX_COMMS_RETRIES; i++) {
    st = fQueryAntennaStatus(verboseBuf);
    if (st != cStNoResponse) {
      break;
    }
  }  /* for */

  /* Return standard status explanation string. */
  if (buf) {
    strcpy(buf, fDecodeAntennaStatus(st, /*maxLength=>*/ 16));
  }
  return ((st == cStTracking) ? 1 : 0);
}  /* fsOnsource */
